<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Definition\ColumnMappingDefinition;
use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Doctrine\DBAL\Exception;
use LogicException;
use Psr\Log\LogLevel;

class Writer extends AbstractSqlProcessor implements WriterInterface
{
    public function write(TableMappingDefinition $tableMappingDefinition, $data): void
    {
        $destQueryBuilder = $this->connection->createQueryBuilder();

        $existingRows = $this->handleSearch($tableMappingDefinition, $data);

        if (null !== $existingRows) {
            $destQueryBuilder->update($tableMappingDefinition->getTargetTableName());

            foreach ($existingRows as $key => $existingRow) {
                $destQueryBuilder
                    ->andWhere("{$existingRow['dest_column_name']} = :{$existingRow['dest_column_name']}{$key}")
                    ->setParameter($existingRow['dest_column_name'] . $key, $existingRow['dest_id']);
            }
        } else {
            $destQueryBuilder
                ->insert($tableMappingDefinition->getTargetTableName());
        }

        $resolvedData = $data;

        foreach ($tableMappingDefinition->getColumnMappingDefinitions() as $columnMappingDefinition) {
            $function = null !== $existingRows ? 'set' : 'setValue';

            call_user_func(
                [$destQueryBuilder, $function],
                $columnMappingDefinition->getTargetColumName(),
                sprintf(':%s', $columnMappingDefinition->getTargetColumName())
            );

            $resolvedValue = $this->resolveMappingValue($columnMappingDefinition, $data);

            $destQueryBuilder
                ->setParameter($columnMappingDefinition->getTargetColumName(), $resolvedValue);

            $resolvedData[$columnMappingDefinition->getTargetColumName()] = $resolvedValue;
        }

        $this->logger->debug($tableMappingDefinition, $destQueryBuilder->getSQL(), $data);

        if (!$this->config->isCommit()) {
            return;
        }

        $destQueryBuilder->execute();
        $id = $this->connection->lastInsertId();

        if (null === $existingRows) {
            $resolvedData['id'] = $id;
            $this->writeTmpTable($tableMappingDefinition, $resolvedData);
        }
    }

    public function writeTmpTable(TableMappingDefinition $tableMappingDefinition, array $data): void
    {
        foreach ($tableMappingDefinition->getIdentifierMappingDefinitions() as $identifierMappingDefinition) {
            $qb = $this->connection->createQueryBuilder();
            $qb->insert($this->config->getTempMappingTableName())
                ->setValue('source_table_name', ':source_table_name')
                ->setValue('source_column_name', ':source_column_name')
                ->setValue('dest_table_name', ':dest_table_name')
                ->setValue('dest_column_name', ':dest_column_name')
                ->setValue('source_id', ':source_id')
                ->setValue('dest_id', ':dest_id')
                ->setParameter('source_table_name', $identifierMappingDefinition->getSourceTableName())
                ->setParameter('source_column_name', $identifierMappingDefinition->getSourceColumnName())
                ->setParameter('dest_table_name', $identifierMappingDefinition->getDestTableName())
                ->setParameter('dest_column_name', $identifierMappingDefinition->getDestColumnName())
                ->setParameter('source_id', $data[$identifierMappingDefinition->getSourceColumnName()])
                ->setParameter('dest_id', $data[$identifierMappingDefinition->getDestColumnName()])
            ;

            $qb->execute();
        }
    }

    /**
     * @throws \Doctrine\DBAL\Driver\Exception
     * @throws Exception
     */
    private function resolveMappingValue(ColumnMappingDefinition $columnMappingDefinition, array $data)
    {
        $value = $data[$columnMappingDefinition->getSourceColumnName()] ?? null;

        if (null !== $columnMappingDefinition->getValue()) {
            return $columnMappingDefinition->getValue();
        }

        if (null !== $columnMappingDefinition->getAlias()) {
            $value = $data[$columnMappingDefinition->getAlias()] ?? null;
        }

        if (null !== $columnMappingDefinition->getIdentifierMapping()) {
            $value = $this->getIdentifierMappingValue($columnMappingDefinition->getIdentifierMapping(), $columnMappingDefinition->getSourceColumnName(), $data);
        }

        if (null !== $columnMappingDefinition->getSubQuerySql()) {
            $value = $this->connection->executeQuery($columnMappingDefinition->getSubQuerySql(), $data)->fetchOne();

            if (false === $value) {
                $value = null;
            }
        }

        if (null !== $columnMappingDefinition->getTransformer()) {
            $transformer = $this->transformerRegistry->getTransformer($columnMappingDefinition->getTransformer());
            $value = $transformer->execute($value);
        }

        return $value;
    }

    /**
     * @throws Exception
     */
    private function handleSearch(TableMappingDefinition $tableMappingDefinition, array $data): ?array
    {
        if (!$tableMappingDefinition->hasSearchMapping()) {
            return null;
        }

        $results = [];

        foreach ($tableMappingDefinition->getSearchMapping() as $searchMappingDefinition) {
            $value = $this->prepareDataForSearch($searchMappingDefinition->getType(), $data[$searchMappingDefinition->getValueColum()]);

            $searchQueryBuilder = $this->connection->createQueryBuilder();
            $searchQueryBuilder
                ->select('*')
                ->from($this->config->getTempMappingTableName())
                ->andWhere('dest_table_name = :destTable')
                ->andWhere('source_table_name = :sourceTable')
                ->andWhere('source_column_name = :sourceColumn')
                ->andWhere('source_id = :sourceId')
                ->setParameter('destTable', $tableMappingDefinition->getTargetTableName())
                ->setParameter('sourceTable', $searchMappingDefinition->getSourceTable())
                ->setParameter('sourceColumn', $searchMappingDefinition->getValueColum())
                ->setParameter('sourceId', $value)
                ->setMaxResults(1)
            ;

            $result = $searchQueryBuilder->execute()->fetchAssociative();

            if (false === $result) {
                return null;
            }

            $results[] = $result;
        }

        return $results;
    }

    private function resolveSearchType(string $type): string
    {
        switch ($type) {
            case 'comparison':
                return '%s = :%s';
            case 'not_comparison':
                return '%s != :%s';
            case 'like':
                return '%s LIKE :%s';
            case 'is_null':
                return '%s IS NULL';
            default:
                throw new LogicException('Unsupported type %s', $type);
        }
    }

    private function prepareDataForSearch(string $type, $value)
    {
        switch ($type) {
            case 'comparison':
                return $value;
            case 'like':
                return sprintf('%%%s%%', $value);
            default:
                throw new LogicException('Unsupported type %s', $type);
        }
    }

    private function getIdentifierMappingValue(array $identifierMapping, string $sourceColumnName, array $data)
    {
        $qb = $this->connection->createQueryBuilder();

        $qb->select('dest_id')
            ->from($this->config->getTempMappingTableName())
            ->andWhere('source_table_name = :sourceTable')
            ->andWhere('source_column_name = :sourceColumn')
            ->andWhere('source_id = :sourceId')
            ->setParameter('sourceTable', $identifierMapping['source_table'])
            ->setParameter('sourceColumn', $identifierMapping['source_column'])
            ->setParameter('sourceId', $data[$sourceColumnName])
        ;

        $result = $qb->execute()->fetchOne();

        if (false === $result) {
            return null;
        }

        return $result;
    }

    public function afterMigrate(TableMappingDefinition $tableMappingDefinition): void
    {
        $afterMigrateDefinition = $tableMappingDefinition->getAfterMigrateDefinition();

        if ($afterMigrateDefinition === null) {
            return;
        }

        foreach ($afterMigrateDefinition->getSql() as $sql) {
            $this->logger->log(LogLevel::DEBUG, $sql);
            if ($this->config->isCommit()) {
                $this->connection->executeQuery($sql);
            }
        }
    }
}
