<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Transformer;

use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Enum\MediaDownloadMethodEnum;
use Dedi\SyliusMigrationPlugin\Exception\MediaDownloadException;
use Symfony\Component\Filesystem\Filesystem;

class PathToSyliusImage implements TransformerInterface
{
    private MigrationConfig $config;
    private string $projectDir;
    private ?string $htpasswd;
    private Filesystem $filesystem;

    public function __construct(
        MigrationConfig $config,
        Filesystem $filesystem,
        string $projectDir,
        ?string $htpasswd
    ) {
        $this->config = $config;
        $this->projectDir = $projectDir;
        $this->htpasswd = $htpasswd;
        $this->filesystem = $filesystem;
    }

    public function getType(): string
    {
        return 'path-to-sylius-image';
    }

    public function execute($data)
    {
        $sslChunk = $this->config->isGenerateRelativePathForImages() ? '' : 'https://';
        $urlChunk = $this->config->isGenerateRelativePathForImages() ? '' : $this->config->getNewSiteUrl();

        $originalPath = $data;

        $data = str_replace([
            'https://',
            'http://',
            $this->config->getLegacySiteUrl(),
            $this->config->getLegacySiteMediaPath(),
        ], [
            $sslChunk,
            $sslChunk,
            $urlChunk,
            '',
        ], $data);

        if ($this->config->isGenerateRelativePathForImages()) {
            $data = trim($data, \DIRECTORY_SEPARATOR);
        }

        if ($this->config->getMediaDownloadMethod() === MediaDownloadMethodEnum::ON_MIGRATE) {
            $targetPath = $this->getTargetDir() . $data;
            if (!is_file($targetPath)) {
                $this->writeImage($originalPath, $targetPath);
            }
        }

        return $data;
    }

    private function getTargetDir(): string
    {
        $targetDirectory = $this->projectDir . \DIRECTORY_SEPARATOR . 'public' . \DIRECTORY_SEPARATOR . $this->config->getNewSiteMediaPath();
        if ($this->config->isGenerateRelativePathForImages()) {
            $targetDirectory .= \DIRECTORY_SEPARATOR;
        }

        return $targetDirectory;
    }

    private function writeImage(string $sourcePath, string $targetPath): void
    {
        $parts = parse_url($sourcePath);
        if (false === $parts || !array_key_exists('scheme', $parts)) {
            return;
        }

        $path_parts = array_map('rawurldecode', explode('/', $parts['path']));

        if (null !== $this->htpasswd) {
            $parts['host'] = $this->htpasswd . '@' . $parts['host'];
        }

        $sourcePath = $parts['scheme'] . '://' . $parts['host'] . implode('/', array_map('rawurlencode', $path_parts));

        $content = @file_get_contents($sourcePath);
        if (false === $content) {
            throw new MediaDownloadException(sprintf('Cannot download target image %s', $sourcePath));
        }
        $this->filesystem->dumpFile($targetPath, $content);
    }
}
